using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;

/// <summary>

/// </summary>


public class VaginaSoftPhysics : MVRScript
{
	DAZPhysicsMesh lowerMesh;
	DAZPhysicsMesh innerGenMesh;
	
	GameObject innerGenContainer;
	
	DAZCharacterRun characterRun;
	
	DAZPhysicsMeshSoftVerticesGroup genGroup;
	
	DAZPhysicsMeshSoftVerticesGroup leftGroup;
	DAZPhysicsMeshSoftVerticesGroup rightGroup;
	
	bool isSessionPlugin = false;
	
	Atom mainAtom;
	
	/*
		Small wrapper class to get into the protected members of DAZPhysicsMesh.
	*/
	public class DAZPhysicsMeshWrapper : DAZPhysicsMesh {
		public void InitializeInternal(DAZPhysicsMesh[] ignore) {
			this._hardVerticesGroups = new List<DAZPhysicsMeshHardVerticesGroup>();
			this._softVerticesGroups = new List<DAZPhysicsMeshSoftVerticesGroup>();
			this._colliderGroups = new List<DAZPhysicsMeshColliderGroup>();
			this.ignorePhysicsMeshes = ignore;
		}
	}
	
	public override void Init() {
		// Might be a better way to do this, but works for now
		isSessionPlugin = containingAtom.name == "CoreControl";
		
		
		AddLabel("generalSettings", "General Settings");
		
		AddSlider("Collider Radius", 0.003f, 0.001f, 0.01f, (float value) => {
			if(leftGroup != null) leftGroup.colliderRadius = value;
			if(rightGroup != null) rightGroup.colliderRadius = value;
		});
		
		AddSlider("Link Spring", 100f, 0f, 1000f, (float value) => {
			if(leftGroup != null) leftGroup.linkSpring = value;
			if(rightGroup != null) rightGroup.linkSpring = value;
		},true);
	
	
		// Up and Down
		AddLabel("upAndDownLabel", "Up And Down Movement");
		AddSlider("Up Limit", 0.01f, 0f, 0.1f, (float value) => {
			if(leftGroup != null) leftGroup.normalDistanceLimit = value;
			if(rightGroup != null) rightGroup.normalDistanceLimit = value;
		});
		
		AddSlider("Down Limit", 0.0f, 0f, 0.1f, (float value) => {
			if(leftGroup != null) leftGroup.normalNegativeDistanceLimit = value;
			if(rightGroup != null) rightGroup.normalNegativeDistanceLimit = value;
		});
		
		AddSlider("Joint Spring", 10.0f, 0f, 100.0f, (float value) => {
			if(leftGroup != null) leftGroup.jointSpringNormal = value;
			if(rightGroup != null) rightGroup.jointSpringNormal = value;
		},true);
		
		AddSlider("Joint Damper", 2.0f, 0f, 100.0f, (float value) => {
			if(leftGroup != null) leftGroup.jointDamperNormal = value;
			if(rightGroup != null) rightGroup.jointDamperNormal = value;
		},true);
		
		
		
		// Left To Right

		AddLabel("leftToRight", "Left To Right Movement");
		
		AddSlider("'Out' Limit", 0.03f, 0f, 0.1f, (float value) => {
			if(leftGroup != null) leftGroup.tangent2NegativeDistanceLimit = value;
			if(rightGroup != null) rightGroup.tangent2DistanceLimit = value;
		});
		
		AddSlider("'In' Limit (Best left at 0)", 0.0f, 0f, 0.3f, (float value) => {
			if(leftGroup != null) leftGroup.tangent2DistanceLimit = value;
			if(rightGroup != null) rightGroup.tangent2NegativeDistanceLimit = value;
		});
		
		AddSlider("Joint Spring", 10.0f, 0f, 100.0f, (float value) => {
			if(leftGroup != null) leftGroup.jointSpringTangent2 = value;
			if(rightGroup != null) rightGroup.jointSpringTangent2 = value;
		},true);
		
		AddSlider("Joint Damper", 2.0f, 0f, 100.0f, (float value) => {
			if(leftGroup != null) leftGroup.jointDamperTangent2 = value;
			if(rightGroup != null) rightGroup.jointDamperTangent2 = value;
		},true);
		
		
		
		// Front and Back

		AddLabel("frontAndBackMovement", "Back and Forward Movement");
		AddSlider("Forward Limit", 0.03f, 0f, 0.1f, (float value) => {
			if(leftGroup != null) leftGroup.tangentNegativeDistanceLimit = value;
			if(rightGroup != null) rightGroup.tangentNegativeDistanceLimit = value;
		});
		
		AddSlider("Back Limit", 0.03f, 0f, 0.1f, (float value) => {
			if(leftGroup != null) leftGroup.tangentDistanceLimit = value;
			if(rightGroup != null) rightGroup.tangentDistanceLimit = value;
		});
		
		AddSlider("Joint Spring", 10.0f, 0f, 100.0f, (float value) => {
			if(leftGroup != null) leftGroup.jointSpringTangent = value;
			if(rightGroup != null) rightGroup.jointSpringTangent = value;
		},true);
		
		AddSlider("Joint Damper", 2.0f, 0f, 100.0f, (float value) => {
			if(leftGroup != null) leftGroup.jointDamperTangent = value;
			if(rightGroup != null) rightGroup.jointDamperTangent = value;
		},true);
		
		
		StartCoroutine("InitSoftPhysics");
	}
	
	private void AddLabel(string name, string text, bool rightSide = false){
		var textField = CreateTextField(new JSONStorableString(name, text), rightSide);
		textField.height = 1;	
	}
		


	public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, bool rightSide = false){
		JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
		storableFloat.storeType = JSONStorableParam.StoreType.Full;
		UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
		uiSlider.rangeAdjustEnabled = rangeAdjust;
		RegisterFloat(storableFloat);
		
		if(onChange != null){
			storableFloat.setCallbackFunction = onChange;
		}
		
		return storableFloat;
	}
	
	private IEnumerator InitSoftPhysics(){
		if(isSessionPlugin){
			// Check to see if the prefab already exists (there's a person in the scene)
			GameObject personPrefab = SuperController.singleton.GetCachedPrefab("a_per", "Person");
			if(personPrefab == null){
				// No people in the scene yet. Load one (without adding to scene) and register it as the prefab. 
				// This might have unintended consequences. 
				SuperController.LogMessage("Person prefab not cached. Loading from asset bundle.");
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_per", "Person", typeof(GameObject));
				if (request == null)
				{
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
				yield return base.StartCoroutine(request);

				personPrefab = request.GetAsset<GameObject>();
				if (personPrefab != null)
				{
					SuperController.singleton.RegisterPrefab("a_per", "Person", personPrefab);
					mainAtom = personPrefab.GetComponent<Atom>();
				} else {
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
			} else {
				mainAtom = personPrefab.GetComponent<Atom>();
			}
		} else {
			mainAtom = containingAtom;
		}
		
		GameObject femaleMorphers = mainAtom.transform.Find("rescale2/geometry/FemaleMorphers").gameObject;
		if(femaleMorphers == null){
			SuperController.LogError($"Could not find FemaleMorphers");
			yield break;	
		}
		
		innerGenContainer = GetOrCreateChildObject(femaleMorphers, "InnerGenPhysicsMesh");
		characterRun = mainAtom.GetComponentInChildren<DAZCharacterRun>();
		
		foreach(DAZPhysicsMesh mesh in mainAtom.GetComponentsInChildren<DAZPhysicsMesh>()){
			if(mesh.name.Contains("Lower")){
				foreach(DAZPhysicsMeshSoftVerticesGroup vertGroup in mesh.softVerticesGroups){
					
					if(vertGroup.name == "gen"){
						lowerMesh = mesh;
						genGroup = vertGroup;
					}
					
					// The existing soft physics groups influence some vertices that we want 
					// to add physics for, so we're going remove them
					foreach(var vertSet in vertGroup.softVerticesSets){
						List<int> toRemove = new List<int>();
						for(int i = 0; i < vertSet.influenceVertices.Length; i++){
							int index = vertSet.influenceVertices[i];
							if(index >= 22086 && index <= 22197){
								toRemove.Add(index);
							}
							
							if(index >= 22676 && index <= 22788){
								toRemove.Add(index);
							}
						}
						foreach(int index in toRemove){
							vertSet.RemoveInfluenceVertex(index);
						}
					}
				}
			}
			
		}
		innerGenMesh = innerGenContainer.GetComponent<DAZPhysicsMesh>();
		if(innerGenMesh == null){
			innerGenMesh = innerGenContainer.AddComponent<DAZPhysicsMeshWrapper>();
			
			(innerGenMesh as DAZPhysicsMeshWrapper).InitializeInternal(mainAtom.GetComponentsInChildren<DAZPhysicsMesh>());
			
			var earlyUpdate = innerGenContainer.AddComponent<DAZPhysicsMeshEarlyUpdate>();
			earlyUpdate.dazPhysicsMesh = innerGenMesh;
		
			// Add some new pysics groups
			leftGroup = InitGroup(innerGenMesh, "genLeft", 22086, 22197, lowerMesh.skin.dazMesh);
			rightGroup = InitGroup(innerGenMesh, "genRight", 22676, 22788, lowerMesh.skin.dazMesh);
			
			if(characterRun != null){
				var newMeshes = characterRun.physicsMeshes.ToList();
				newMeshes.Add(innerGenMesh);
				characterRun.physicsMeshes = newMeshes.ToArray();
			}
			
			innerGenMesh.useThreading = true;
			
			innerGenMesh.characterRun = characterRun;
			innerGenMesh.skinTransform = lowerMesh.skin.transform;
			innerGenMesh.skin = lowerMesh.skin;

			innerGenMesh.enabled = true;
			innerGenMesh.isEnabled = true;
			innerGenMesh.updateEnabled = true;
		} else {
			foreach(var vertGroup in innerGenMesh.softVerticesGroups){
				if(vertGroup.name == "genLeft"){
					leftGroup = vertGroup;
				} else if (vertGroup.name == "genRight") {
					rightGroup = vertGroup;
				}
			}
		}
		
		foreach(var entry in floatParams){
			JSONStorableFloat floatParam = entry.Value;
			if(floatParam.setCallbackFunction != null)
				floatParam.setCallbackFunction(floatParam.val);
		}
		
		innerGenContainer.SetActive(true);
		
		SuperController.LogMessage($"Successfully added new soft physics groups");
	}
	
	public DAZPhysicsMeshSoftVerticesGroup InitGroup(DAZPhysicsMesh parent, string name, int startVertIndex, int endVertIndex, DAZMesh dazMesh){
		int newGroupIndex = parent.AddSoftVerticesGroup();
		var newGroup = parent.softVerticesGroups[newGroupIndex];
		
		newGroup.name = name;
		
		// For now we'll only create groups that influence one vertex. 
		// This is more compute-intensive but the vaginal canal doesn't
		// have too many vertices
		for(int i = startVertIndex; i <= endVertIndex; i++){
			int newSetIndex = newGroup.AddSet();
			var newSet = newGroup.softVerticesSets[newSetIndex];
			newSet.targetVertex = i;
			newSet.anchorVertex = i;
			newSet.influenceVertices = new int[]{};
		}
		
		foreach(var vertSet in newGroup.softVerticesSets){
			var setDistances = closestToVertex(vertSet, dazMesh.baseMesh.vertices, newGroup.softVerticesSets, 4);
			foreach(var setDistance in setDistances){
				if(vertSet.anchorVertex == vertSet.targetVertex){
					vertSet.anchorVertex = setDistance.set.targetVertex;
				}
				
				var linkSet = setDistance.set;
				
				bool alreadyLinked = false;
				foreach(var otherLink in linkSet.links){
					if(otherLink == vertSet.uid){
						alreadyLinked = true;
						break;
					}
				}
				if(!alreadyLinked){
					vertSet.links.Add(linkSet.uid);
				}
			}
		}
		
		newGroup.normalReference = genGroup.normalReference;
		
		
		
		newGroup.ignoreColliders = genGroup.ignoreColliders;
		newGroup.ignoreAutoColliderGroups = genGroup.ignoreAutoColliderGroups;
		newGroup.otherGroupNumsCollisionAllowed  = new int[0];
		newGroup.colliderMaterial = genGroup.colliderMaterial;
		
		newGroup.colliderType = DAZPhysicsMeshSoftVerticesGroup.ColliderType.Sphere;
		newGroup.centerBetweenTargetAndAnchor = false;
		newGroup.lookAtOption = DAZPhysicsMeshSoftVerticesGroup.LookAtOption.NormalReference;
		newGroup.colliderOrient = DAZPhysicsMeshSoftVerticesGroup.ColliderOrient.Tangent;
		
		newGroup.normalMovementType = DAZPhysicsMeshSoftVerticesGroup.MovementType.Limit;
		newGroup.tangentMovementType = DAZPhysicsMeshSoftVerticesGroup.MovementType.Limit;
		newGroup.tangent2MovementType = DAZPhysicsMeshSoftVerticesGroup.MovementType.Limit;
		
		
		newGroup.useThreading = true;
		newGroup.useParentSettings = false;
		
		return newGroup;
	}
	
	public class VertSetDistance {
		public DAZPhysicsMeshSoftVerticesSet set;
		public float distance;
	}
	
	public LinkedList<VertSetDistance> closestToVertex(DAZPhysicsMeshSoftVerticesSet baseSet, Vector3[] vertices, List<DAZPhysicsMeshSoftVerticesSet> sets, int max){
		LinkedList<VertSetDistance> list = new LinkedList<VertSetDistance>();
		
		Vector3 baseVertex = vertices[baseSet.targetVertex];
		
		foreach(var otherSet in sets){
			if(baseSet == otherSet) continue;
			
			Vector3 otherVertex = vertices[otherSet.targetVertex];
			
			float distance = (baseVertex - otherVertex).magnitude;
			VertSetDistance entry = new VertSetDistance();
			entry.set = otherSet;
			entry.distance = distance;
			
			InsertSorted(list, entry, max);
		}
		
		return list;
	}
	
	public static void InsertSorted(LinkedList<VertSetDistance> list, VertSetDistance value, int maxSize)
	{
		var newNode = new LinkedListNode<VertSetDistance>(value);

		if (list.First == null || value.distance < list.First.Value.distance)
		{
			list.AddFirst(newNode);
			return;
		}

		var current = list.First;
		while (current.Next != null && current.Next.Value.distance < value.distance)
		{
			current = current.Next;
		}

		list.AddAfter(current, newNode);
		
		if (list.Count > maxSize)
		{
			while (list.Count > maxSize)
			{
				list.RemoveLast();
			}
		}
	}
	
	private void Start()
	{
	}
	
	private void OnEnable(){
		if(innerGenMesh != null){
			innerGenMesh.isEnabled = true;
			innerGenMesh.collisionEnabled = true;
		}
	}
	
	private void OnDisable()
	{
		if(innerGenMesh != null){
			innerGenMesh.isEnabled = false;
			innerGenMesh.collisionEnabled = false;
		}
	}
	
	private void OnDestroy(){
		if(characterRun != null){
			var newMeshes = characterRun.physicsMeshes.ToList();
			newMeshes.Remove(innerGenMesh);
			characterRun.physicsMeshes = newMeshes.ToArray();
		}
		
		if(innerGenContainer != null){
			innerGenContainer.name += "Deleted";
			Destroy(innerGenContainer);
		}
	}
	
	
	private GameObject FindGameObject(GameObject parent, string name){
		Transform childTransform = parent.transform.Find(name);
		
		if(childTransform == null){
			return null;
		} else {
			return childTransform.gameObject;
		}
	}
	
	private GameObject GetOrCreateChildObject(GameObject parent, string name) {
		GameObject gameObject = FindGameObject(parent, name);
		if(gameObject == null) {
			gameObject = new GameObject(name);
			gameObject.transform.parent = parent.transform;
		} 
		
		return gameObject;
	}
}
